/*
* Copyright 2018 Adrian Panazzolo <panaaj@hotmail.com>
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

const pkg= require('../package.json')
const fs= require('fs')
const path= require('path')
const geoJSON = require('geojson-validation')
const uuid = require('uuid/v4')

module.exports= function (app) {

    let plugin= {
        id: 'gpxload',
        name: 'GPXLoad: Resource Provider',
        description: pkg.description,
        version: pkg.version
    }
    let config= {
        API: {
            put: true,
            delete: true,
            delta: false
        },
        activeRoute: {
            href: null,
            startTime: null
        },
        nextPoint: { position: null }       
    }

    let unsubscribe= []     // stream subscriptions
    let timers= []          // interval imers

    const uuidPrefix= 'urn:mrn:signalk:uuid:'
    const apiUri='/api'
    const manageUri='/manage'
    const savePath= app.config.configPath + '/resources'
    const resTypes= {
        route: { path: path.join(savePath, '/routes')},
        waypoint: { path: path.join(savePath, '/waypoints')},
        track: { path: path.join(savePath, '/tracks')}
    }
    
    plugin.start= (props)=> {
        try {
            app.debug(`${plugin.name} starting.......`) 
            app.debug(`** pkgname: ${pkg.name}, pkgversion: ${pkg.version}`) 

            config= props || config;
            if(!config.activeRoute) { 
                config.activeRoute= {
                    href: null,
                    startTime: null
                }
            }
            if(!config.nextPoint) { 
                config.nextPoint= { position: null }
            }            
            app.debug('*** Configuration ***')
            app.debug(config)           

            checkSavePath()
            .then( res=> {
                let ae= (config.API && config.API.put) ? 'PUT enabled, ' : 'PUT disabled, ';
				ae+= (config.API && config.API.delete) ? 'DELETE enabled' : 'DELETE disabled';
                app.setProviderStatus(`Started. ${ae}.`) 
                app.debug(`** ${plugin.name} started... ${(res) ? 'OK' : 'with errors!'}`)     
            })
            .catch( e=> { app.debug(e) } )
            
            // **register HTTP PUT handlers
            if(app.registerActionHandler) {
                app.debug('** Registering Action Handler(s) **')

                app.registerActionHandler(
                    'vessels.self',
                    'resources.routes',
                    handlePutResource
                )  
                app.registerActionHandler(
                    'vessels.self',
                    'resources.waypoints',
                    handlePutResource
                )      
                app.registerActionHandler(
                    'vessels.self',
                    'navigation.courseGreatCircle.activeRoute.href',
                    handlePutCourseData
                )     
                app.registerActionHandler(
                    'vessels.self',
                    'navigation.courseGreatCircle.activeRoute.startTime',
                    handlePutCourseData
                )   
                app.registerActionHandler(
                    'vessels.self',
                    'navigation.courseGreatCircle.nextPoint.position',
                    handlePutCourseData
                )                  
            } 

            // ** register STREAM UPDATE message handlers
            app.debug('** Registering STREAM UPDATE Handler(s) **')

            unsubscribe.push(   // ** handle route Update
                app.streambundle.getSelfBus('resources.routes')
                .onValue( v=> { handleResourceUpdate(v) })
            )
            unsubscribe.push(   // ** handle waypoint Update
                app.streambundle.getSelfBus('resources.waypoints')
                .onValue( v=> { handleResourceUpdate(v) })
            )    
            unsubscribe.push( // ** handle activeRoute.href Update
                app.streambundle.getSelfBus('navigation.courseGreatCircle.activeRoute.href')
                .onValue( v=> {     
                    if(v['$source']==plugin.id) { return }
                    setActiveRoute(v.value)                  
                })
            )
            unsubscribe.push( // ** handle activeRoute.startTime Update
                app.streambundle.getSelfBus('navigation.courseGreatCircle.activeRoute.startTime')
                .onValue( v=> {     
                    if(v['$source']==plugin.id) { return }
                    setActiveRoute(v.value)     
                })
            )   
            unsubscribe.push( 
                app.streambundle.getSelfBus('navigation.courseGreatCircle.nextPoint.position')
                .onValue( v=> {     
                    if(v['$source']==plugin.id) { return }
                    setNextPoint(v.value)    
                })
            )   
            
            afterStart()
        } 
        catch (e) {
            app.setProviderError(`Started with errors!`)
            app.error("error: " + e)
            console.error(e.stack)
            return e
        }       
    }

    plugin.stop= ()=> { 
        app.debug(`${plugin.name} stopping.......`) 
        app.debug('** Un-registering Update Handler(s) **')
        unsubscribe.forEach( b=> b() )
        unsubscribe= []
        app.debug('** Stopping Timer(s) **')
        timers.forEach( t=> clearInterval(t) )
        timers= []        
        app.setProviderStatus(`Stopped`)
    }

    plugin.schema= { 
        properties: { 
            API: {
                type: "object",
                description: `Use this section to allow 
                    applications to use the Signal K HTTP / STREAM APIs to update,
                    replace and delete resources.`,
                properties: {
                    put: {
                        type: "boolean", 
                        title: "Enabled"
                    },
                    delta: {
                        type: "boolean",
                        title: "Enabled"
                    },                   
                    delete: {
                        type: "boolean",
                        title: "Enabled"
                    }                    
                }
            }
        } 
    }

    plugin.uiSchema= {
        API: {
            put: {
                "ui:widget": "checkbox",
                "ui:title": "Enable / Disable PUT API",
                "ui:help": "Check this box to allow resources to be added / updated via PUT api"
            },
            delete: {
                "ui:widget": "checkbox",
                "ui:title": "Allow DELETE",
                "ui:help": "Check this box to allow resources to be deleted via HTTP / STREAM apis"
            },
            delta: {
                "ui:widget": "checkbox",
                "ui:title": "Allow update via Delta",
                "ui:help": "Check this box to allow resources to be added, updated and deleted via Delta messages"
            }                          
        }
    }

    plugin.registerWithRouter= router=> {
        // ** accept submitted resources **
        router.post(apiUri, (req, res) => {
            if(!Array.isArray(req.body)) {
                res.status(401)
                res.send({error: true, message: `${plugin.id}: Invalid request data received!`})
            }
            else {
                processUIPost(req.body)
                res.status(200)
                res.send({error: false, message: `${plugin.id}: OK`})
            }
        })

        // ** return persisted resources **
        router.get(apiUri, (req, res) => {
            try {
                res.status(200)
                res.json( getPersistedResources() )
            }
            catch(err)  {
                res.status(401)
                res.send({error: true, message: `${plugin.id}: Error retrieving resources!`})
            }
        }) 

        // ** manage resources **
        router.post(manageUri, (req, res) => {
            if(!req.body.action) {
                res.status(401)
                res.send({error: true, message: `${plugin.id}: Invalid request data received!`})
            }
            else {
                switch(req.body.action) {
                    case 'activeRoute':
                        let now= new Date()
                        let dt= req.body.value ? now.toISOString() : null
                        config.activeRoute.href= req.body.value
                        config.activeRoute.startTime= dt
                        let val= [
                            {path: req.body.path, value: req.body.value},
                            {path: 'navigation.courseGreatCircle.activeRoute.startTime', value: dt }
                        ]
                        app.debug(`****** Sending Delta: ******`)
                        app.debug(JSON.stringify({updates: [ {values: val} ] }))
                        app.handleMessage(plugin.id, {updates: [ {values: val} ] })
                        break
                    case 'delete':
                        let r={}
                        r.type= req.body.value.type.slice(0,req.body.value.type.length-1)
                        r.id= req.body.value.value.uuid
                        r.value= null    
                        app.debug(r)
                        if(persist(r)) { 
                            return { 
                                state: 'COMPLETED', 
                                resultStatus: 200, 
                                statusCode: 200 
                            } 
                        }
                        else {
                            return { 
                                state: 'COMPLETED', 
                                resultStatus: 502, 
                                statusCode: 502,
                                message: `Invalid resource data values!` 
                            }                    
                        } 
                        break                        
                    default:
                        res.status(401)
                        res.send({error: true, message: `${plugin.id}: Invalid request data received!`})
                }
                res.status(200)
                res.send({error: false, message: `${plugin.id}: OK`})
            }
        })        
        
        // ** return plugin information **
        router.get('/about', (req, res) => {
            app.debug(req)
            res.status(200)
            res.send({
                id: plugin.id,
                name: plugin.name,
                version: plugin.version,
                description: plugin.description,
                url: pkg.name
            }) 
        })  
    }   

    plugin.signalKApiRoutes= router=> {
        router.get('/resources/waypoints/meta', (req, res)=> {
            res.json( {description: 'Collection of waypoints, each named with a UUID'} )
        })        
        router.get('/resources/waypoints', (req, res)=> {
            res.json( getPersistedResources('waypoint').waypoints )
        })   
        router.get(`/resources/waypoints/${uuidPrefix}*-*-*-*-*`, (req, res)=> {
            let r= getHttpData(req, 'waypoint') 
            if(r.error) { res.send( r.message ) }
            else { res.json(r.data) }
        })  
        router.post(`/resources/waypoints`, (req, res)=> {
            let path= req.path.slice(1).split('/').join('.')
            let value= {}
            let id= uuidPrefix + uuid()
            value[id]= (typeof req.body.value!=='undefined') ? req.body.value : req.body
            res.json( handlePutResource(null, path, value, null) )
        })  
        router.delete(`/resources/waypoints/${uuidPrefix}*-*-*-*-*`, (req, res)=> {
            let p= req.path.slice(1).split('/')
            let path= p.slice(0,2).join('.')               
            let id= p[2]
            let value= {}
            value[id]= null
            res.json( handlePutResource(null, path, value, null) )
        })                   

        router.get('/resources/routes/meta', (req, res)=> {
            res.json( {description: 'Collection of routes, each named with a UUID'} )
        })  
        router.get('/resources/routes', (req, res) => {
            res.json( getPersistedResources('route').routes )
        })   
        router.get(`/resources/routes/${uuidPrefix}*-*-*-*-*`, (req, res)=> {
            let r= getHttpData(req, 'route') 
            if(r.error) { res.send( r.message ) }
            else { res.json(r.data) }
        }) 
        router.post(`/resources/routes`, (req, res)=> {
            let path= req.path.slice(1).split('/').join('.')
            let value= {}
            let id= uuidPrefix + uuid()
            value[id]= (typeof req.body.value!=='undefined') ? req.body.value : req.body
            res.json( handlePutResource(null, path, value, null) )
        }) 
        router.delete(`/resources/routes/${uuidPrefix}*-*-*-*-*`, (req, res)=> {
            let p= req.path.slice(1).split('/')
            let path= p.slice(0,2).join('.')               
            let id= p[2]
            let value= {}
            value[id]= null
            res.json( handlePutResource(null, path, value, null) )
        })        
     
        router.get('/resources/tracks', (req, res)=> {
            res.json( getPersistedResources('track').tracks )
        })  
        router.get(`/resources/tracks/${uuidPrefix}*-*-*-*-*`, (req, res)=> {
            let r= getHttpData(req, 'track') 
            if(r.error) { res.send( r.message ) }
            else { res.json(r.data) }
        })  
     
        return router
    }

    // ****************************************

    // ** additional plugin start processing 
    function afterStart() {
        app.debug('** afterStart() **')
        // ** Freshen persisted COURSE data in UPDATE cache
        emitCourseData()
        timers.push( setInterval( emitCourseData, 30000 ) )
    }

    // ** emit delta UPDATE message for persisted COURSE data 
    function emitCourseData() {
        // ** send initial values delta **
        let val= []
        if(config.activeRoute.href) {
            val.push({
                path: 'navigation.courseGreatCircle.activeRoute.href', 
                value: config.activeRoute.href
            })
        }
        if(config.activeRoute.startTime) {
            val.push({
                path: 'navigation.courseGreatCircle.activeRoute.startTime', 
                value: config.activeRoute.startTime
            })
        }
        if(config.nextPoint.position) {
            val.push({
                path: 'navigation.courseGreatCircle.nextPoint.position', 
                value: config.nextPoint.position
            })                                 
        }
        if(val.length!=0) {
            app.debug(`****** Emit COURSE data: ******`)
            app.debug(val)
            app.handleMessage(plugin.id, {updates: [ {values: val} ] })  
        }
    }

    // *** Course data processing ************************************

    function handlePutCourseData(context, path, value, cb) {
        app.debug(` 
            ${JSON.stringify(path)}, 
            ${JSON.stringify(value)}`
        )  
        let ok= false
        let val= [ {path: path, value: value} ]

        let p= path.split('.')
        if(p[2]=='activeRoute') {
            if(p[p.length-1]=='href') { 
                ok= setActiveRoute(value, 'href')
                let st= (value) ? config.activeRoute.startTime : null 
                val.push( {
                    path: 'navigation.courseGreatCircle.activeRoute.startTime', 
                    value: st
                })
            }
            if(p[p.length-1]=='startTime') { ok= setActiveRoute(value, 'startTime') }
        }        
        if(p[2]=='nextPoint') {
            if(p[p.length-1]=='position') { ok= setNextPoint(value) }
        }      
        if(ok) {
            // ** send delta **
            app.debug(`****** Sending Delta: ******`)
            app.debug(val)
            app.handleMessage(plugin.id, {updates: [ {values: val} ] })   
            return { state: 'COMPLETED', resultStatus: 200, statusCode: 200 } 
        }  
        else {
            return { 
                state: 'COMPLETED', 
                resultStatus: 400, 
                statusCode: 400,
                message: `Invalid resource reference!` 
            }              
        } 
    }

    function setActiveRoute(value, key) {
        app.debug('** setActiveRoute **')
        app.debug(key)
        app.debug(value)

        if(key=='href') { 
            config.activeRoute.href= value 
            let dt= new Date()
            let st= (value) ? dt.toISOString() : null
            config.activeRoute.startTime= st
        }
        if(key=='startTime') { config.activeRoute.startTime= value }
        app.savePluginOptions(config, res=> {app.debug('config saved.')})   
             
        return true   
    }    

    function setNextPoint(value) {
        app.debug('** setNextPoint **')
        app.debug(value)

        if(value) {
            if(typeof value.latitude === 'undefined' || 
                typeof value.longitude === 'undefined') { return false }
        }
        config.nextPoint.position= value
        app.savePluginOptions(config, res=> {app.debug('config saved.')})     

        return true
    }     

    // *** resource processing **************************************

    // ** compile http api get response **
    function getHttpData(req, resType) {
        let res= {error: false, message: '', data: null }

        let p= parsePath(req.path)
        if(!p.path) { 
            res.error= true
            res.message= `Cannot GET ${req.path}`
            return res
        }

        let fname= p.path.substring(p.path.lastIndexOf(':')+1)
        let r= getPersistedResources(resType, fname)
        
        if(r.error) { 
            res.error= true
            res.message= `Cannot GET ${req.path}`
            return res
        }

        if(p.attribute) { 
            let a= eval(`r.${p.attribute}`)
            if(a) { res.data= a }
            else { 
                res.error= true
                res.message= `Cannot GET ${req.path}` 
            }
        }
        else { res.data=r }

        return res         
    }

    // ** parse provided path to resource and attributes
    function parsePath(path) {
        let res= {
            path: null,
            attribute: null
        }
        let a= path.split('/')
        if( isUUID(a[a.length-1]) ) { res.path= path }
        else {  
            if( isUUID(a[3]) ) {
                res.path= a.slice(0,4).join('/')
                res.attribute= a.slice(4).join('.')
            }
        }
        app.debug(res)
        return res
    }

    // ** check / create path to persist resources
    function checkSavePath() {
        return new Promise( (resolve, reject)=> {
            fs.access(savePath, fs.constants.W_OK | fs.constants.R_OK, (err)=>{
                if(err) {
                    app.debug(`${savePath} not detected.... attempting to create it....`)
                    fs.mkdir(savePath, (err)=> {
                        if(err) { 
                            raiseSavePathError(savePath) 
                            resolve(false)
                        }   
                        else {
                            app.debug(`** OK: ${savePath} created....`)
                            let rtype= Object.keys(resTypes)
                            rtype.forEach( t=> {
                                fs.mkdir(resTypes[t].path, (err)=> {
                                    if(err) { raiseSavePathError(p) }  
                                    else { app.debug(`** OK: ${resTypes[t].path} created....`) }                           
                                })
                            })
                            resolve(true)                       
                        }
                    })
                }
                else { resolve(true) }
            })
        })
    }

    // ** raise save path error
    function raiseSavePathError(path) {
        let p= `Unable to create ${path} folder`
        app.error(`*** ERROR: ${p} ***`)
        app.setProviderError(p)
    }

    //** return persisted resources from storage
    function getPersistedResources(type=null, item=null) {
        app.debug('*** Retrieving saved Resource(s) ***')
        let result= {}
        let rTypes= Object.entries(resTypes);
        try{
            if(item) {
                result= JSON.parse(fs.readFileSync( path.join(resTypes[type].path, item) , 'utf8'))
				app.debug(`Returning resource: ${item}`)
            }
            else {
                rTypes.forEach(rt=> {         
                    if(!type || type==rt[0]) {
                        result[rt[0] + 's']= {}
                        let files= fs.readdirSync(rt[1].path)
                        files.forEach( f=> {
                            let uuid= uuidPrefix + f
							try {
								result[rt[0] + 's'][uuid]= JSON.parse(fs.readFileSync( path.join(rt[1].path, f) , 'utf8'))
							}
							catch(err) {
								app.debug(`Invalid file contents: ${f}`)
							}
                        })                
                    }
                })  
            }  
            return result
        }
        catch(err) {
            app.debug(`Resource NOT Found: ${item}`)
            return {
                error: true, 
                message: `Error retreiving resources from ${savePath}. Ensure plugin is active or restart plugin!`,
                source: plugin.id
            }
        }
    }    

    // ** process posted data **
    function processUIPost(data) {
        app.debug(`** Loading resources.........`)
        if(!Array.isArray(data)) {
            app.debug(`** ERROR: ** Invalid payload data!!`)
            return
        }
        data.forEach( r=> { persist(r) })
    }   
    
    // ** handle Resource PUT via registerActionHandler
    function handlePutResource(context, path, value, cb) {
        app.debug(` 
            ${JSON.stringify(path)}, 
            ${JSON.stringify(value)}`
        )       
        // ** test for allowed PUT api use
        if(!config.API.put) { 
            return { 
                state: 'COMPLETED', 
                resultStatus: 405, 
                statusCode: 405,
                message: `Method Not Allowed!` 
            }            
        }
        let p= path.split('.')
        if(p.length<2) {
            return { 
                state: 'COMPLETED', 
                resultStatus: 400, 
                statusCode: 400,
                message: `Invalid path!` 
            }
        }
        
        let r={}
        r.type= p[1].slice(0,p[1].length-1)
        let v= Object.entries(value)
        r.id= v[0][0]
        r.value= v[0][1]
        app.debug(r)

        // ** test for allowed DELETE via PUT api
        if(r.value===null && !config.API.delete) { 
            return { 
                state: 'COMPLETED', 
                resultStatus: 405, 
                statusCode: 405,
                message: `Method Not Allowed!` 
            }            
        }        
        // ** test for valid resource identifier
        if( !isUUID(r.id) ) {
            return { 
                state: 'COMPLETED', 
                resultStatus: 400, 
                statusCode: 400,
                message: `Invalid resource id!` 
            }
        }        

        switch(r.type) {
            case 'route': 
            case 'waypoint':
            case 'track':
                if(persist(r)) { 
                    return { state: 'COMPLETED', resultStatus: 200, statusCode: 200 } 
                }
                else {
                    return { 
                        state: 'COMPLETED', 
                        resultStatus: 502, 
                        statusCode: 502,
                        message: `Invalid resource data values!` 
                    }                    
                }           
                break
            default:
                return { 
                    state: 'COMPLETED', 
                    resultStatus: 400, 
                    statusCode: 400,
                    message: `Invalid resource type (${r.type})!` 
                }
        }
    }    

    // ** handle resource Update via streambundle.getSelfBus
    function handleResourceUpdate(data) {
        if(data['$source']==plugin.id) { return }
                   
        // ** test for allowed use (use PUT api config)
        if(!config.API.delta) { 
            app.debug('Delta Updates are disabled!')
            return 
        }
        
        let r={}
        let p= data.path.split('.');
        r.type= p[1].slice(0,p[1].length-1)
        let v= Object.entries(data.value)
        r.id= v[0][0]
        r.value= v[0][1]
        app.debug('Data to persist...')
        app.debug(r)

        // ** test for allowed DELETE (use DELETE api config)
        if(r.value===null && !config.API.delete) { 
            app.debug('DELETE via Delta Updates are disabled!')
            return 
        }    

        // ** test for valid resource identifier
        if( !isUUID(r.id) ) { return }        

        switch(r.type) {
            case 'route': 
            case 'waypoint':
            case 'track':
                persist(r)
        }
    }       

    // ** save / delete (r.value==null) resource file
    function persist(r) {
        if( !isUUID(r.id) ) { return false }
        let fname= r.id.substring(r.id.lastIndexOf(':')+1)
        let p= path.join(resTypes[r.type].path, fname)
        let action= (r.value===null) ? 'DELETE' : 'SAVE'
        app.debug(`******  ${r.type}: ${action} -> ${fname} ******`)
        if(r.value===null) { // ** delete file **
            if(r.type=='route') { // for route delete start & end waypoints
                // get start / end waypoint uuids
                let w= getPersistedResources('route', fname)
                if(w.start) {
                    app.debug(`** ${action} Start waypoint: ${w.start} **`)
                    let wf= w.start.substring(w.start.lastIndexOf(':')+1)
                    let wp= path.join(resTypes['waypoint'].path, wf)
                    fs.unlink(wp, err=> { 
                        if(err) { app.debug('Error deleting file!') }
                        else { sendDelta( {id: w.start, type: 'waypoint', value: null}) }
                    })  
                }
                if(w.end) {
                    app.debug(`** ${action} End waypoint: ${w.end} **`)
                    wf= w.end.substring(w.end.lastIndexOf(':')+1)
                    wp= path.join(resTypes['waypoint'].path, wf)
                    fs.unlink(wp, err=> { 
                        if(err) { app.debug('Error deleting file!') } 
                        else { sendDelta( {id: w.end, type: 'waypoint', value: null}) }
                    })
                }
            }
            fs.unlink(p, err=> {
                if(err) { app.debug('Error deleting file!') }
                else { app.debug('File deleted!')}
            })
            sendDelta(r)

        }
        else {  // ** add / update file
            if( !validateData(r) ) { return false }
            // ** test for valid SignalK value **
            fs.writeFile(p, JSON.stringify(r.value), (err, res)=> {
                if(err) { app.debug(err) }
                else { app.debug(`** ${r.type} written to ${fname} **`) }
            }) 
            sendDelta(r)
            if(r.type=='route' && Array.isArray(r.waypoints) ) { 
                r.waypoints.forEach( w=> { persist(w) }) 
            }  
        }
        return true;
    }  

    // ** send delta message for resource
    function sendDelta(r) {
        let key= r.id
        let p= `resources.${r.type}s.${key}`
        let val= [{path: p, value: r.value}]
        app.debug(`****** Send Delta: ******`)
        app.debug(JSON.stringify({updates: [ {values: val} ] }))
        app.handleMessage(plugin.id, {updates: [ {values: val} ] })
    }

    // ** validate provided resource value data
    function validateData(r) {
        if(!r.type) { return false }
        switch(r.type) {
            case 'route':
                return validateRoute(r.value);
                break;
            case 'waypoint':
                return validateWaypoint(r.value);
                break;
            case 'track':
                return validateTrack(r.value);
                break;   
            default:
                return false;             
        }      
    }  
    
    // ** validate route data
    function validateRoute(r) {   
        if(typeof r.name == 'undefined') { return false }
        if(typeof r.description == 'undefined') { return false }
        if(typeof r.distance == 'undefined' || isNaN(r.distance)) { return false }
        if(!r.start) { return false }
        if(!r.end) { return false }
        if(!r.feature || !geoJSON.valid(r.feature)) { 
            app.debug(`****** INVALID GeoJSON  ******`)
            return false 
        }
        app.debug(`****** VALID Route  ******`)
        return true
    } 

    // ** validate waypoint data
    function validateWaypoint(r) {  
        if(!r.position) { return false } 
        if(!r.position.latitude || !r.position.longitude) { return false } 
        if(!r.feature || !geoJSON.valid(r.feature)) { 
            app.debug(`****** INVALID GeoJSON  ******`)
            return false 
        }
        app.debug(`****** VALID Waypoint  ******`)
        return true
    } 

    // ** validate track data
    function validateTrack(r) {  return true  } 

    // ** returns true if id is a valid UUID **
    function isUUID(id) {
        let uuid= RegExp("^urn:mrn:signalk:uuid:[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-4[0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$")
        let result= uuid.test(id)
        if(!result) { app.debug(`****** INVALID ID: ${id} ******`) }
        return result
    }

    return plugin
}


